from cmd import Cmd
from hca_util.hca_util import *
from hca_util.aws import *

h_config="""usage: config ACCESS_KEY SECRET_KEY
\tConfigure your machine with credentials"""

h_create="""usage: create [project_name] [-udx]
\tCreate an upload directory for project (authorised user only)
\tIf specified, project name needs to be between 1-36 alphanumeric characters with no space
\tIf specified, allowed permissions include 'u', 'ud', 'ux' and 'udx'; otherwise default 'ux'
\tu - upload, d - download, x - delete"""

h_list="""usage: list
\tList contents of bucket (authorised user only)
usage: list DIR_NAME
\tList contents of directory"""

h_select="""usage: select DIR_NAME
\tSet active directory for upload and download"""

h_dir="""usage: dir
\tShow selected directory"""

h_upload="""usage: upload F1 [f2] [f3] ...
\tMulti-files upload to selected directory
usage: upload .
\tUpload all files from current user directory"""

h_delete="""usage: delete F1 [f2] [f3] ...
\tDelete specified file(s) from selected directory
usage: delete .
\tDelete all files from selected directory
usage: delete
\tDelete selected directory (authorised user only)"""

h_download="""usage: download F1 [f2] [f3] ...
\tDownload specified file(s) from selected directory to local machine
usage: download .
\tDownload all files from selected directory to local machine"""

h_exit="""usage: exit (or quit)
\tExit the tool. Shorthand: x, q, or Ctrl-D"""


class HcaCmd(Cmd):
    prompt = 'hca> '
    intro = 'Type ? to list commands'

    def __init__(self, args):
        super().__init__() # call parent class constructor

        profile=Aws.DEFAULT_PROFILE
        region=Aws.DEFAULT_REGION

        if args.profile is not None:
            profile = args.profile
            print('Using profile: ' + profile)

        if args.region is not None:
            region = args.region
            print('Using region: ' + region)

        self.util = HcaUtil(profile, region)

    # commands
    def do_config(self, inp):
        self.util.cmd_config(inp.split())

    def do_create(self, inp):
        self.util.cmd_create(inp.split())

    def do_list(self, inp):
        self.util.cmd_list(inp.split())

    def do_select(self, inp):
        self.util.cmd_select(inp.split())

    def do_dir(self, inp):
        self.util.cmd_dir(inp.split())

    def do_upload(self, inp):
        self.util.cmd_upload(inp.split())

    def do_delete(self, inp):
        self.util.cmd_delete(inp.split())

    def do_download(self, inp):
        self.util.cmd_download(inp.split())

    def do_exit(self, inp):
        print('Bye')
        return True

    def default(self, inp):
        if inp == 'x' or inp == 'q' or inp == 'quit':
            return self.do_exit(inp)

        print(f'No such command. Try `help`')

    # help info
    def help_config(self):
        print(h_config)

    def help_create(self):
        print(h_create)

    def help_list(self):
        print(h_list)

    def help_select(self):
        print(h_select)

    def help_dir(self):
        print(h_dir)

    def help_upload(self):
        print(h_upload)

    def help_delete(self):
        print(h_delete)

    def help_download(self):
        print(h_download)

    def help_exit(self):
        print(h_exit)


    do_EOF = do_exit
    help_EOF = help_exit




class LocalState:

    def __init__(self):
        self.known_dirs = []
        self.selected_dir = None

    def add_dir(self, dir):
        if dir not in self.known_dirs:
            self.known_dirs.append(dir)

    def del_dir(self, dir):
        if dir in self.known_dirs:
            self.known_dirs.remove(dir)

    def reset_dirs(self):
        self.known_dirs = []

    def select_dir(self, dir):
        self.selected_dir = dir
        self.addDir(dir)

    def unselect_dir(self):
        self.selected_dir = None



#!/usr/bin/python

import os
from multiprocessing.dummy import Pool

import boto3
from botocore.exceptions import ClientError

from hca_util.aws import *
from hca_util.bucket_policies import *
from hca_util.common import *
from hca_util.upload_progress import UploadProgress


class HcaUtil:
    session = None
    aws = None
    bucket_name = None
    setup_ok = False

    selected_dir = None

    # default constructor
    def __init__(self, profile, region):
        self.profile = profile
        self.region = region
        self.setup()

    # setup function
    def setup(self):
        try:
            # try to set a session using profile_name
            if not Aws.profile_exists(self.profile):
                print(f'Profile \'{self.profile}\' not found')
            else:
                profile_dict = Aws.get_profile(self.profile)
                self.session = boto3.Session(region_name=profile_dict['region'],
                                             aws_access_key_id=profile_dict['access_key'],
                                             aws_secret_access_key=profile_dict['secret_key'])

                # use profile to create clients/resources for aws services: s3, secret_mgr, sts
                self.aws = Aws(self.session)

                # get bucket name from aws secrets (also serve as a way to validate user config/creds)
                try:
                    # access policy can't be attached to a secret
                    # GetSecretValue action should be allowed for user
                    resp = self.aws.secret_mgr.get_secret_value(SecretId=Aws.SECRET_NAME)
                    secret_str = resp['SecretString']
                    self.bucket_name = json.loads(secret_str)['s3-bucket']
                except:
                    print(f'Invalid credentials')

                if self.bucket_name:
                    self.setup_ok = True
        except Exception as e:
            print(f'An exception of type {e.__class__.__name__} occurred in setup.\nDetail: ' + str(e))

        if not self.setup_ok:
            print('See `help config` for help to configure your credentials')

    # can't proceed with most common if setup nok
    # except: config, dir

    # command functions
    def cmd_config(self, argv):
        """
        we may not have a session yet, if setup() wasn't successful.
        or we may have.
        In any case, we have to call setup() again after we have new config/creds.
        :param argv:
        :return:
        """

        if len(argv) == 2:
            access_key = argv[0]
            secret_key = argv[1]
            try:
                Aws.set_profile(self.profile, self.region, access_key, secret_key)
                self.setup()

            except Exception as e:
                print(f'An exception of type {e.__class__.__name__} occurred in cmd config.\nDetail: ' + str(e))
        else:
            print('Invalid args. See `help config`')

    def cmd_create(self, argv):
        if not self.setup_ok:
            print(f'Setup failed: \nSee `help config` for help to configure your credentials')
            return

        if len(argv) > 2:
            print('Invalid args. See `help create`')
            return

        # generate random uuid prefix for directory name
        dir_name = gen_uuid()

        # valid input
        # 1. create
        # 2. create [-udx]
        # 3. create [project_name]
        # 4. create [project_name] [-udx]

        def verify_perms(permissions):
            is_valid_perms = permissions in allowed_perms_combinations
            if not is_valid_perms:
                permissions = default_perms
                print('Invalid perms, using default')
            print(f'Perms <{permissions}>')
            return permissions

        def verify_projname(proj_name):
            is_valid_projname = is_valid_project_name(proj_name)
            if not is_valid_projname:
                proj_name = ''
                print('Invalid project name, ignoring')
            print(f'Project name <{proj_name}>')
            return proj_name

        project_name = ''
        perms = default_perms
        if len(argv) == 0:
            print(f'Project name <>')
            print(f'Default perms <{perms}>')
        elif len(argv) == 1 and argv[0].startswith('-'):
            print(f'Project name <>')
            perms = verify_perms(argv[0][1:])
        elif len(argv) == 1 and not argv[0].startswith('-'):
            project_name = verify_projname(argv[0])
            print(f'Default perms <{perms}>')
        elif len(argv) == 2 and argv[1].startswith('-'):
            project_name = verify_projname(argv[0])
            perms = verify_perms(argv[1][1:])
        else:
            print('Invalid args. See `help create`')
            return

        try:

            self.aws.s3.put_object(Bucket=self.bucket_name, Key=(dir_name + '/'), Metadata={'name': f'{project_name}'})
            print('Created ' + dir_name)

            # get bucket policy
            s3_resource = self.session.resource('s3')
            try:
                bucket_policy = s3_resource.BucketPolicy(self.bucket_name)
                policy_str = bucket_policy.policy
            except ClientError:
                policy_str = ''

            if policy_str:
                policy_json = json.loads(policy_str)
            else:  # no bucket policy
                policy_json = json.loads('{ "Version": "2012-10-17", "Statement": [] }')

            # add new statement for dir to existing bucket policy
            new_statement = new_policy_statement(self.bucket_name, dir_name, perms)
            policy_json['Statement'].append(new_statement)

            updated_policy = json.dumps(policy_json)

            bucket_policy.put(Policy=updated_policy)

        except Exception as e:
            print(f'An exception of type {e.__class__.__name__} occurred in cmd create.\nDetail: ' + str(e))

    def cmd_list(self, argv):
        if not self.setup_ok:
            print(f'Setup failed: \nSee `help config` for help to configure your credentials')
            return

        if len(argv) == 0 or len(argv) == 1:

            try:
                s3_resource = self.aws.session.resource('s3')
                bucket = s3_resource.Bucket(self.bucket_name)

                if len(argv) == 0:

                    for obj in bucket.objects.all():
                        k = obj.key
                        if k.endswith('/'):
                            n = obj.Object().metadata.get('name')
                            print(k + (f' {n}' if n else ''))
                else:
                    dir_name = argv[0]
                    if is_valid_dir_name(dir_name):

                        dir_name += '' if dir_name.endswith('/') else '/'

                        for obj in bucket.objects.filter(Prefix=dir_name):
                            k = obj.key
                            if not k.endswith('/'):
                                print(k)

                    else:
                        print('Invalid directory name')
            except Exception as e:
                print(f'An exception of type {e.__class__.__name__} occurred in cmd list.\nDetail: ' + str(e))

        else:
            print('Invalid args. See `help list`')

    def cmd_select(self, argv):
        if not self.setup_ok:
            print(f'Setup failed: \nSee `help config` for help to configure your credentials')
            return

        if len(argv) == 1:
            dir_name = argv[0]
            if is_valid_dir_name(dir_name):

                dir_name += '' if dir_name.endswith('/') else '/'

                try:
                    s3_resource = self.aws.session.resource('s3')
                    bucket = s3_resource.Bucket(self.bucket_name)
                    bucket.Object(dir_name)
                    # serialize(select_dir, dir_name)
                    self.selected_dir = dir_name
                    print('Selected ' + dir_name)

                except Exception as e:
                    print(f'An exception of type {e.__class__.__name__} occurred in cmd select.\nDetail: ' + str(e))

            else:
                print('Invalid directory name')

        else:
            print('Invalid args. See `help select`')

    def cmd_dir(self, argv):
        """Returns currently selected dir or None."""
        print(self.selected_dir)

    def cmd_upload(self, argv):
        if not self.setup_ok:
            print(f'Setup failed: \nSee `help config` for help to configure your credentials')
            return

        if len(argv) > 0:

            if self.selected_dir is None:
                print('No directory selected')
            else:
                if len(argv) == 1 and argv[0] == '.':
                    # upload files from current directory
                    fs = [f for f in os.listdir('.') if os.path.isfile(f)]
                else:
                    # upload specified list of files
                    fs = [f for f in argv if os.path.exists(f)]

                print('Uploading...')

                try:
                    with Pool(12) as p:
                        p.map(lambda f: self.upload(f), fs)
                        print('Done.')

                except Exception as e:
                    print(f'An exception of type {e.__class__.__name__} occurred in cmd upload.\nDetail: ' + str(e))

        else:
            print('Invalid args. See `help upload`')

    """
    It is recommended to create a resource instance for each thread / process in a multithreaded or 
    multiprocess application rather than sharing a single instance among the threads / processes
    """

    def upload(self, f):
        # upload_file automatically handles multipart uploads via the S3 Transfer Manager
        # put_object maps to the low-level S3 API request, it does not handle multipart uploads
        self.aws.session.resource('s3').Bucket(self.bucket_name) \
            .upload_file(Filename=f, Key=self.selected_dir + os.path.basename(f),
                         Callback=UploadProgress(f))

    def cmd_delete(self, argv):
        if not self.setup_ok:
            print(f'Setup failed: \nSee `help config` for help to configure your credentials')
            return

        if self.selected_dir is None:
            print('No directory selected')
            return

        try:
            prefix = self.selected_dir
            s3_resource = self.aws.session.resource('s3')
            bucket = s3_resource.Bucket(self.bucket_name)

            if len(argv) == 0:
                # delete an entire directory
                confirm = input(f'Confirm delete {self.selected_dir} and its content? Y/y to proceed: ')
                if confirm == 'Y' or confirm == 'y':
                    print('Deleting...')

                    for obj in bucket.objects.filter(Prefix=prefix):
                        print('Deleting ' + obj.key)
                        obj.delete()

                    # reset selected dir
                    self.selected_dir = None
                    print('Selected dir: None')

                else:
                    print('Delete cancelled')

                return

            if len(argv) == 1 and argv[0] == '.':
                # delete all files in selected directory
                for obj in bucket.objects.filter(Prefix=prefix):
                    # do not delete folder object
                    if obj.key == prefix:
                        continue
                    print('Deleting ' + obj.key)
                    obj.delete()
            else:
                # delete specified file(s) in selected directory
                for f in argv:
                    print('Deleting ' + prefix + f)
                    obj = bucket.Object(prefix + f)
                    obj.delete()

        except Exception as e:
            print(f'An exception of type {e.__class__.__name__} occurred in cmd delete.\nDetail: ' + str(e))

    def cmd_download(self, argv):
        if not self.setup_ok:
            print(f'Setup failed: \nSee `help config` for help to configure your credentials')
            return

        if self.selected_dir is None:
            print('No directory selected')
            return

        if len(argv) < 1:
            print('Invalid args. See `help download`')
            return

        try:
            prefix = self.selected_dir
            s3_resource = self.aws.session.resource('s3')
            bucket = s3_resource.Bucket(self.bucket_name)
            if len(argv) == 1 and argv[0] == '.':
                # download all files from selected directory
                for obj in bucket.objects.filter(Prefix=prefix):
                    # do not download folder object
                    if obj.key == prefix:
                        continue
                    print('Downloading ' + obj.key)
                    if not os.path.exists(os.path.dirname(obj.key)):
                        os.makedirs(os.path.dirname(obj.key))
                    bucket.download_file(obj.key, obj.key)
            else:
                # download specified file(s) from selected directory
                for f in argv:
                    print('Downloading ' + prefix + f)
                    obj = bucket.Object(prefix + f)
                    if not os.path.exists(os.path.dirname(obj.key)):
                        os.makedirs(os.path.dirname(obj.key))
                    obj.download_file(obj.key)

        except Exception as e:
            print(f'An exception of type {e.__class__.__name__} occurred in cmd config.\nDetail: ' + str(e))



import configparser
from pathlib import Path
import json


class Aws:
    HOME = str(Path.home())
    CONFIG_FILE = HOME + '/.aws/config'
    CREDENTIALS_FILE = HOME + '/.aws/credentials'

    # default profile use creds from [DEFAULT_PROFILE] section of ~/.aws/credentials
    # and config from [profile DEFAULT_PROFILE] section of ~/.aws/config
    DEFAULT_PROFILE = 'hca-util'
    DEFAULT_REGION = 'us-east-1'
    SECRET_NAME = 'hca/util/secret'

    def __init__(self, session):
        self.session = session
        self.s3 = session.client('s3')
        self.secret_mgr = session.client('secretsmanager')
        self.sts = session.client('sts')

    def sts_get_caller_arn(self):
        try:
            resp = self.sts.get_caller_identity()
            arn = resp.get('Arn')
            return arn
        except Exception as e:
            print(str(e))
            sys.exit(1)

    @staticmethod
    def profile_exists(profile):
        # let's not bother checking CONFIG_FILE to see if region is set
        # we can always use default region
        credentials = configparser.ConfigParser(comment_prefixes='/', allow_no_value=True)
        credentials.read(Aws.CREDENTIALS_FILE)

        if credentials.has_section(profile):
            return True

        return False

    @staticmethod
    def get_profile(profile):
        credentials = configparser.ConfigParser()
        credentials.read(Aws.CREDENTIALS_FILE)

        profile_dict = {'access_key': '',
                        'secret_key': '',
                        'region': ''}

        if credentials.has_section(profile):
            profile_dict['access_key'] = credentials[profile].get('aws_access_key_id')
            profile_dict['secret_key'] = credentials[profile].get('aws_secret_access_key')

        config = configparser.ConfigParser()
        config.read(Aws.CONFIG_FILE)

        if config.has_section(f'profile {profile}'):
            profile_dict['region'] = config[f'profile {profile}'].get('region')

        if not profile_dict['region']:
            profile_dict['region'] = Aws.DEFAULT_REGION

        return profile_dict

    @staticmethod
    def set_profile(profile, region, access_key, secret_key):

        """.aws/config
        [profile {profile}]
        region = {region}
        """

        # set comment_prefixes to a string which you will not use in the config file
        config = configparser.ConfigParser(comment_prefixes='/', allow_no_value=True)
        config.read(Aws.CONFIG_FILE)

        if not config.has_section(f'profile {profile}'):
            config.add_section(f'profile {profile}')
        config.set(f'profile {profile}', 'region', region)

        with open(Aws.CONFIG_FILE, 'w') as out:
            config.write(out)

        """.aws/credentials
        [{profile}]
        aws_access_key_id = {0}
        aws_secret_access_key = {1}
        """

        credentials = configparser.ConfigParser(comment_prefixes='/', allow_no_value=True)
        credentials.read(Aws.CREDENTIALS_FILE)

        if not credentials.has_section(f'{profile}'):
            credentials.add_section(f'{profile}')
        credentials.set(f'{profile}', 'aws_access_key_id', access_key)
        credentials.set(f'{profile}', 'aws_secret_access_key', secret_key)

        with open(Aws.CREDENTIALS_FILE, 'w') as out:
            credentials.write(out)

        print('Credentials saved.')



# __main__.py

import argparse
from hca_util.hca_cmd import HcaCmd
from hca_util.aws import *


def main():
    parser = argparse.ArgumentParser(description='hca-util')
    parser.add_argument(
        '--profile',
        help=f'use PROFILE instead of default \'{Aws.DEFAULT_PROFILE}\' profile'
    )
    parser.add_argument(
        '--region',
        help=f'use REGION instead of default \'{Aws.DEFAULT_REGION}\' region'
    )
    args = parser.parse_args()
    HcaCmd(args).cmdloop()


if __name__ == '__main__':
    main()



# default = ux

# u - upload
# d - download
# x - delete

allowed_perms_combinations = ['u', 'ud', 'ux', 'udx']
default_perms = 'ux'

s3_permissions = {'u': 's3:PutObject',
                  'd': 's3:GetObject',
                  'x': 's3:DeleteObject'}


def get_policy_statement_template():
    return {
        'Action': [],
        'Effect': 'Allow',
        'Resource': 'arn:aws:s3:::BUCKET_NAME/DIR_NAME/*',
        'Principal': {
            'AWS': [
                'arn:aws:iam::871979166454:user/HCAContributor'
            ]
        }
    }.copy()


def new_policy_statement(bucket, dir, perms=None):
    statement = get_policy_statement_template()
    # add permissions/actions
    if perms is None or perms not in allowed_perms_combinations:
        perms = default_perms
    for p in perms:
        statement['Action'].append(s3_permissions.get(p))
        res = statement['Resource']
        statement['Resource'] = res.replace('BUCKET_NAME', bucket).replace('DIR_NAME', dir)
    return statement



import os
import pathlib
from setuptools import setup

# directory containing this file
HERE = pathlib.Path(__file__).parent

# text of the README file
README = (HERE / "README.md").read_text()

# install requirements
INSTALL_REQS = [line.rstrip() for line in open(os.path.join(os.path.dirname(__file__), "requirements.txt"))]

# This call to setup() does all the work
setup(
    # dashes are ok in repo and PyPI dist names but not in package (i.e. directory) and
    # module (.py file) names. can't do import xyz-abc
    name="hca-util",
    version="0.0.15",
    description="CLI tool to allow submitters to upload to and download from the EBI S3 buckets.",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/ebi-ait/hca-util",
    author="hca-ingest-dev",
    author_email="prabhat@ebi.ac.uk",
    license="Apache License",
    classifiers=[
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: POSIX",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
    ],
    platforms=['MacOS X', 'Posix'],
    packages=["hca_util"],
    include_package_data=True,
    install_requires=INSTALL_REQS,
    entry_points={
        "console_scripts": [
            "hca-util=hca_util.__main__:main",
        ]
    },
)
